/************************************************************************\
*                                                                        *
*  Descripcin: Funciones de acceso a OLE desde FW                       *
*                                                                        *
*  Autor: Jos F. Gimnez (JFG) - <jfgimenez@wanadoo.es>                 *
*                                 <tecnico.sireinsa@ctv.es>              *
*  Fecha: 6-11-2000                                                      *
*                                                                        *
*  Revisado: 13-2-2001, 28-2-2001, 13-3-2001, 18-5-2001                  *
*                                                                        *
*  Nota: basado en las funciones de acceso a OLE de FW 2.1               *
*                                                                        *
*        Compilar via assembler para evitar unresolved external:         *
*        FIWRQQ, FIDRQQ, etc. en Clipper 5.2                             *
*                                                                        *
**************************************************************************
*                                                                        *
*  CreateOleObject( cOleName | cCLSID [, cIID ] ) -> hOleObject          *
*                                                                        *
*                                                                        *
*  OleInvoke( hOleObject, cMethodName, uParam1, ..., uParamN )           *
*          -> uResult                                                    *
*                                                                        *
*                                                                        *
*  OleSetProperty( hOleObject, cPropertyName, uParam1, ..., uParamN )    *
*          -> lOk                                                        *
*                                                                        *
*                                                                        *
*  OleGetProperty( hOleObject, cPropertyName, uParam1, ..., uParamN )    *
*          -> uResult                                                    *
*                                                                        *
*                                                                        *
*  OleQueryInterface( hOleObject, cIID ) -> ppvObject                    *
*                                                                        *
*                                                                        *
*  OleAddRef( hOleObject ) -> nRefCount                                  *
*                                                                        *
*                                                                        *
*  OleRelease( hOleObject ) -> nRefCount                                 *
*                                                                        *
*                                                                        *
*  OleCOMFunction( hOleObject, nMethod, uParam1, ..., uParamN )          *
*          -> uResult                                                    *
*                                                                        *
*                                                                        *
*  OleIsObject() -> lIsObject                                            *
*                                                                        *
*                                                                        *
*  OleError() -> nError                                                  *
*                                                                        *
*                                                                        *
*  Ole2TxtError() -> cError                                              *
*                                                                        *
*                                                                        *
*  OleUninitialize() -> Nil                                              *
*                                                                        *
*                                                                        *
*  AnsiToWide( cAnsiStr ) -> cWideStr                                    *
*                                                                        *
*                                                                        *
*  WideToAnsi( cWideStr ) -> cAnsiStr                                    *
*                                                                        *
\************************************************************************/


#define _BC5

#ifndef __FLAT__
   #include <WinTen.h>
#endif

#include <Windows.h>
#include <Ole2.h>

#ifndef __FLAT__
   #include <dispatch.h>
   #include <CoGuid.h>
   #include <OleNls.h>
   #include <ClipApi.h>
#else
   #include <OleAuto.h>
   #include <OleDB.h>
   #include <ShlObj.h>
#endif

#ifdef __HARBOUR__
   #define _CLIPDEFS_H
   #include <extend.h>
   #include <HBdate.h>
   #include <HBvm.h>
   #undef  WORD
   #define WORD        unsigned short
   #define NIL         '\0'
   #undef  PCOUNT
   #define PCOUNT()    hb_pcount()
   #define _param      hb_param
   #define _bset       memset
   #define PCLIPVAR    PHB_ITEM
   #define wType       type
   #define vt          n1.n2.vt
   #define u           n1.n2.n3
#endif

// -----------------------------------------------------------------------
// Estas funciones estan declaradas expresamente como PASCAL, aunque en
// realidad son C. Sirven para apilar parametros.

void pascal _pushc( char * cParam );
void pascal _pushl( int lParam );
void pascal _pushnl( int lParam );
void pascal _pushnd( double ndParam );

// -----------------------------------------------------------------------

static far VARIANTARG RetVal;

static EXCEPINFO excep;

static HRESULT nOleError = 0;

static int nInitialized = 0;

//---------------------------------------------------------------------------//

#ifndef __FLAT__
extern long _dAlphaToDateDBF( LPSTR cDate );
extern void _dDateToAlphaDBF( LPSTR cDate, long nDate );
extern long _dvtol( double );
extern void _retds( LPSTR cDate );
#endif

//---------------------------------------------------------------------------//

static double DateToDbl( LPSTR cDate )
{
   double nDate;

#ifndef __FLAT__
   nDate = _dAlphaToDateDBF( cDate ) - 0x0024d9abL;   // 31-12-1899
#else
   nDate = hb_dateEncStr( cDate ) - 0x0024d9abL;
#endif

   return ( nDate );
}

//---------------------------------------------------------------------------//

static LPSTR DblToDate( double nDate )
{
   static char *cDate = "00000000";

#ifndef __FLAT__
   long nlDate = _dvtol( nDate );

   _dDateToAlphaDBF( cDate, nlDate + 0x0024d9abL );
#else
   hb_dateDecStr( cDate, nDate + 0x0024d9abL );
#endif

   return ( cDate );
}

//---------------------------------------------------------------------------//

static LPSTR AnsiToWide( LPSTR cAnsi )
{
   WORD wLen;
   LPSTR cString;

   #ifdef __FLAT__
      wLen  = MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, cAnsi, -1, 0, 0 );
      cString = ( char * ) _xgrab( wLen * 2 );
      MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, cAnsi, -1,
                           ( LPWSTR ) cString, wLen );
   #else
      cString = cAnsi;
   #endif

   return ( cString );
}

//---------------------------------------------------------------------------//

static LPSTR WideToAnsi( LPSTR cWide )
{
   WORD wLen;
   LPSTR cString;

   #ifdef __FLAT__
      wLen = WideCharToMultiByte( CP_ACP, 0, ( LPWSTR ) cWide, -1,
                                  cString, 0, NULL, NULL );
      cString = _xgrab( wLen );
      WideCharToMultiByte( CP_ACP, 0, ( LPWSTR ) cWide, -1,
                           cString, wLen, NULL, NULL );
   #else
      cString = cWide;
   #endif

   return ( cString );
}

//---------------------------------------------------------------------------//

static void GetParams(DISPPARAMS * dParams)
{
   VARIANTARG * pArgs = NULL;
   PCLIPVAR uParam, Self;
   int n, nArgs, nArg;
   double date;
   LPSTR cString;

   nArgs = PCOUNT() - 2;

   if( nArgs > 0 )
   {
      pArgs = ( VARIANTARG * ) _xgrab( sizeof( VARIANTARG ) * nArgs );

      for( n = 0; n < nArgs; n++ )
      {
         // Los parametros en VARIANTARG[] hay que ponerlos en orden inverso
         nArg = nArgs + 2 - n;

         VariantInit( &( pArgs[ n ] ) );

         uParam = _param( nArg, 0xFFFF );

         switch( uParam->wType )
         {
            case NIL:
                 pArgs[ n ].vt   = VT_EMPTY;
                 break;

            case CHARACTER:
            case MEMO:
                 pArgs[ n ].vt   = VT_BSTR;
                 cString = AnsiToWide( _parc( nArg ) );
                 pArgs[ n ].u.bstrVal = SysAllocString( (LPVOID) cString );
                 #ifdef __FLAT__
                    _xfree( cString );
                 #endif
                 break;

            case LOGICAL:
                 pArgs[ n ].vt   = VT_BOOL;
                 #ifdef __FLAT__
                    pArgs[ n ].u.boolVal = _parl( nArg );
                 #else
                    pArgs[ n ].u.bool = _parl( nArg );
                 #endif
                 break;

#ifdef __FLAT__
            case HB_IT_INTEGER:
            case HB_IT_LONG:
#endif
            case NUMERIC:
                 pArgs[ n ].vt   = VT_I4;
                 pArgs[ n ].u.lVal = _parnl( nArg );
                 break;

#ifndef __FLAT__
            case NUM_FLOAT:
                 pArgs[ n ].vt   = VT_R8;
                 uParam = _param( nArg, 0xFFFF);
                 _bcopy( ( LPVOID ) &pArgs[ n ].u.dblVal, ( LPVOID ) &uParam->pPointer1, 8 );
                 break;
#else
            case HB_IT_DOUBLE:
                 pArgs[ n ].vt   = VT_R8;
                 pArgs[ n ].u.dblVal = hb_parnd( nArg );
                 break;
#endif

#ifndef __FLAT__
            case DATE:
#else
            case HB_IT_DATE:
#endif
                 pArgs[ n ].vt   = VT_DATE;
                 pArgs[ n ].u.dblVal = DateToDbl( _pards( nArg ) );
                 break;

#ifndef __FLAT__
            case OBJECT:
                 {
                 PCLIPSYMBOL pData = _Get_Sym( "hObj" );
                 if( pData )
                    {
                    _PutSym( pData );
                    _xPushM( uParam );
                    _xSend( 0 );
                    pArgs[ n ].vt = VT_DISPATCH;
                    pArgs[ n ].u.pdispVal = ( IDispatch * ) _parnl( -1 );
                    }
                 else
                    pArgs[ n ].vt = VT_EMPTY;
                 }
                 break;
#else
            case HB_IT_OBJECT:
                 {
                 PHB_DYNS pData;
                 pArgs[ n ].vt = VT_EMPTY;
                 if ( hb_stricmp( hb_objGetClsName( uParam ), "TOleAuto" ) == 0 )
                    {
                    pData = hb_dynsymFindName( "hObj" );
                    if( pData )
                       {
                       hb_vmPush( uParam );
                       hb_vmMessage( pData->pSymbol );
                       hb_vmDo( 0 );
                       pArgs[ n ].vt = VT_DISPATCH;
                       pArgs[ n ].u.pdispVal = ( IDispatch * ) _parnl( -1 );
                       }
                    }
                 }
                 break;
#endif
         }
      }
   }

   dParams->rgvarg = pArgs;
   dParams->cArgs  = nArgs;
   dParams->rgdispidNamedArgs = 0;
   dParams->cNamedArgs = 0;

}

//---------------------------------------------------------------------------//

static void FreeParams(DISPPARAMS * dParams)
{
   int n;

   if( dParams->cArgs > 0 )
   {
      for( n = 0; n < ( int ) dParams->cArgs; n++ )
         VariantClear( &(dParams->rgvarg[ n ]) );

      _xfree( ( LPVOID ) dParams->rgvarg );
   }
}

//---------------------------------------------------------------------------//

static void RetValue( void )
{
   LPSTR cString;

   switch( RetVal.vt )
   {
      case VT_BSTR:
           cString = WideToAnsi( ( LPSTR ) RetVal.u.bstrVal );
           _retc( cString );
           #ifdef __FLAT__
              _xfree( cString );
           #endif
           break;

      case VT_BOOL:
           #ifdef __FLAT__
              _retl( RetVal.u.boolVal );
           #else
              _retl( RetVal.u.bool );
           #endif
           break;

      case VT_DISPATCH:
           _retnl( ( LONG ) RetVal.u.pdispVal );
           break;

      case VT_I4:
           _retnl( ( LONG ) RetVal.u.iVal );
           break;

      case VT_R8:
           #ifndef __FLAT__
              _bcopy( ( LPVOID ) &_eval->pPointer1, ( LPVOID ) &RetVal.u.dblVal, 8 );
              _eval->wType = 8;
           #else
              hb_retnd( RetVal.u.dblVal );
           #endif
           break;

      case VT_DATE:
           _retds( DblToDate( RetVal.u.dblVal ) );
           break;

      case VT_EMPTY:
           _ret();
           break;

      default:
           if ( nOleError == S_OK )
              (LONG) nOleError = -1;
           _ret();
           break;
   }

   if( RetVal.vt != VT_DISPATCH )
      VariantClear( &RetVal );

}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_CREATEOLEOBJECT() // ( cOleName | cCLSID  [, cIID ] )
#else
CLIPPER CREATEOLEO( PARAMS ) // ( cOleName | cCLSID  [, cIID ] )
#endif
{
   LPSTR cCLSID;
   GUID ClassID, iid;
   REFIID riid = &IID_IDispatch;
   IDispatch * pDisp = NULL;

   nOleError = S_OK;

   if ( nInitialized == 0 )
      nOleError = OleInitialize( NULL );

   if ( (nOleError == S_OK) || (nOleError == (HRESULT) S_FALSE) )
      {
      nInitialized++;

      cCLSID = AnsiToWide( _parc( 1 ) );
      if ( _parc( 1 )[ 0 ] == '{' )
         nOleError = CLSIDFromString( ( LPOLESTR ) cCLSID, &ClassID );
      else
         nOleError = CLSIDFromProgID( ( LPCOLESTR ) cCLSID, &ClassID );
      #ifdef __FLAT__
         _xfree( cCLSID );
      #endif

      if ( PCOUNT() == 2 )
         {
         if ( _parc( 2 )[ 0 ] == '{' )
            {
            cCLSID = AnsiToWide( _parc( 2 ) );
            nOleError = CLSIDFromString( ( LPOLESTR ) cCLSID, &iid );
            #ifdef __FLAT__
               _xfree( cCLSID );
            #endif
            }
         else
            memcpy( ( LPVOID ) &iid, _parc( 2 ), sizeof( iid ) );

         ( LPVOID ) riid = &iid;
         }

      if ( nOleError == S_OK )
         nOleError = CoCreateInstance( &ClassID, NULL, CLSCTX_SERVER,
                                       riid, (LPVOID) &pDisp );
      }

   _retnl( ( LONG ) pDisp );

}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_OLESHOWEXCEPTION()
#else
CLIPPER OLESHOWEXC()
#endif
{
   if ( (LONG) nOleError == DISP_E_EXCEPTION )
      #ifdef __FLAT__
      {
         LPSTR source, description;

         source = WideToAnsi( (LPVOID) excep.bstrSource );
         description = WideToAnsi( (LPVOID) excep.bstrDescription );
         MessageBox( NULL, description, source, MB_ICONHAND );
         _xfree( source );
         _xfree( description );
      }
      #else
         MessageBox( NULL, excep.bstrDescription, excep.bstrSource, MB_ICONHAND );
      #endif
}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_OLEINVOKE() // (hOleObject, szMethodName, uParams...)
#else
CLIPPER OLEINVOKE( PARAMS ) // (hOleObject, szMethodName, uParams...)
#endif
{
   IDispatch * pDisp = ( IDispatch * ) _parnl( 1 );
   LPSTR cMember;
   DISPID lDispID;
   DISPPARAMS dParams;
   UINT uArgErr;

   VariantInit( &RetVal );
   _bset( (LPBYTE) &excep, 0, sizeof( excep ) );

   cMember = AnsiToWide( _parc( 2 ) );
   nOleError = pDisp->lpVtbl->GetIDsOfNames( pDisp, &IID_NULL,
                                             ( LPVOID ) &cMember, 1,
                                             LOCALE_USER_DEFAULT, &lDispID );
   #ifdef __FLAT__
      _xfree( cMember );
   #endif

   if ( nOleError == S_OK )
      {
      GetParams( &dParams );
      nOleError = pDisp->lpVtbl->Invoke( pDisp,
                                         lDispID,
                                         &IID_NULL,
                                         LOCALE_USER_DEFAULT,
                                         DISPATCH_METHOD,
                                         &dParams,
                                         &RetVal,
                                         &excep,
                                         &uArgErr ) ;
      FreeParams( &dParams );
      }

   RetValue();
}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_OLESETPROPERTY() // (hOleObject, cPropName, uValue, uParams...)
#else
CLIPPER OLESETPROP( PARAMS ) // (hOleObject, cPropName, uValue, uParams...)
#endif
{
   IDispatch * pDisp = ( IDispatch * ) _parnl( 1 );
   LPSTR cMember;
   DISPID lDispID, lPropPut = DISPID_PROPERTYPUT;
   DISPPARAMS dParams;
   UINT uArgErr;

   VariantInit( &RetVal );
   _bset( (LPBYTE) &excep, 0, sizeof( excep ) );

   cMember = AnsiToWide( _parc( 2 ) );
   nOleError = pDisp->lpVtbl->GetIDsOfNames( pDisp, &IID_NULL,
                                             ( LPVOID ) &cMember, 1,
                                             LOCALE_USER_DEFAULT, &lDispID );
   #ifdef __FLAT__
      _xfree( cMember );
   #endif

   if ( nOleError == S_OK )
      {
      GetParams( &dParams );
      dParams.rgdispidNamedArgs = &lPropPut;
      dParams.cNamedArgs = 1;

      nOleError = pDisp->lpVtbl->Invoke( pDisp,
                                         lDispID,
                                         &IID_NULL,
                                         LOCALE_USER_DEFAULT,
                                         DISPATCH_PROPERTYPUT,
                                         &dParams,
                                         NULL,    // No return value
                                         &excep,
                                         &uArgErr );

      FreeParams( &dParams );
      }

   _ret();
}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_OLEGETPROPERTY()  // (hOleObject, cPropName, uParams...)
#else
CLIPPER OLEGETPROP( PARAMS )  // (hOleObject, cPropName, uParams...)
#endif
{
   IDispatch * pDisp = ( IDispatch * ) _parnl( 1 );
   LPSTR cMember;
   DISPID lDispID;
   DISPPARAMS dParams;
   UINT uArgErr;

   VariantInit( &RetVal );
   _bset( (LPBYTE) &excep, 0, sizeof( excep ) );

   cMember = AnsiToWide( _parc( 2 ) );
   nOleError = pDisp->lpVtbl->GetIDsOfNames( pDisp, &IID_NULL,
                                             ( LPVOID ) &cMember, 1,
                                             LOCALE_USER_DEFAULT, &lDispID );
   #ifdef __FLAT__
      _xfree( cMember );
   #endif

   if ( nOleError == S_OK )
      {
      GetParams( &dParams );

      nOleError = pDisp->lpVtbl->Invoke( pDisp,
                                         lDispID,
                                         &IID_NULL,
                                         LOCALE_USER_DEFAULT,
                                         DISPATCH_PROPERTYGET,
                                         &dParams,
                                         &RetVal,
                                         &excep,
                                         &uArgErr );

      FreeParams( &dParams );
      }

   RetValue();

}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_OLEQUERYINTERFACE()  // (hOleObject, cIID ) -> ppvObject
#else
CLIPPER OLEQUERYIN( PARAMS )  // (hOleObject, cIID ) -> ppvObject
#endif
{
   IUnknown * pUnk = ( IUnknown * ) _parnl( 1 );
   IUnknown * ppvObject = NULL;
   GUID iid;
   LPSTR ciid;

   nOleError = S_OK;

   if ( _parc( 2 )[ 0 ] == '{' )
      {
      ciid = AnsiToWide( _parc( 2 ) );
      nOleError = CLSIDFromString( ( LPOLESTR ) ciid, &iid );
      #ifdef __FLAT__
         _xfree( ciid );
      #endif
      }
   else
      memcpy( ( LPVOID ) &iid, _parc( 2 ), sizeof( iid ) );

   if ( nOleError == S_OK )
      nOleError = pUnk -> lpVtbl -> QueryInterface( pUnk,
                                                    ( LPVOID ) &iid,
                                                    ( LPVOID ) &ppvObject );

   _retnl( ( LONG ) ppvObject );

}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_OLEADDREF()  // ( hOleObject )
#else
CLIPPER OLEADDREF( PARAMS )  // ( hOleObject )
#endif
{
   IUnknown * pUnk = ( IUnknown * ) _parnl( 1 );

   _retnl( pUnk -> lpVtbl -> AddRef( pUnk ) );
}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_OLERELEASE()  // ( hOleObject )
#else
CLIPPER OLERELEASE( PARAMS )  // ( hOleObject )
#endif
{
   IUnknown * pUnk = ( IUnknown * ) _parnl( 1 );

   _retnl( pUnk -> lpVtbl -> Release( pUnk ) );
}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_COMFUNCTION()  // ( hOleObject, nFunc, uParams... )
{
   typedef HRESULT ( STDMETHODCALLTYPE * COMFunc ) ( IUnknown * pUnk );

   IUnknown * pUnk = ( IUnknown * ) _parnl( 1 );
   COMFunc pFunc;
   COMFunc *vTbl;
   int i;
   double doubles[16];
   LPVOID ptros[16];


   vTbl = ( COMFunc * ) &( pUnk -> lpVtbl -> QueryInterface );
   vTbl += _parni( 2 ) + 2;
   pFunc = *vTbl;

   for ( i = PCOUNT(); i > 2; i-- )
      switch ( (_parinfo( i ) & ~HB_IT_BYREF) )
         {
         case CHARACTER:
         case MEMO:
              _pushc( _parc( i ) );
              break;

         case LOGICAL:
              if ( ISBYREF( i ) )
                 {
                 ptros[ i ] = (LPVOID) _parl( i );
                 _pushc( (char *) &ptros[ i ] );
                 }
              else
                 _pushl( _parl( i ) );
              break;

         case HB_IT_INTEGER:
         case HB_IT_LONG:
              if ( ISBYREF( i ) )
                 {
                 ptros[ i ] = (LPVOID) _parnl( i );
                 _pushc( (char *) &ptros[ i ] );
                 }
              else
                 _pushnl( _parnl( i ) );
              break;

         case HB_IT_DOUBLE:
              if ( ISBYREF( i ) )
                 {
                 doubles[ i ] = _parnd( i );
                 _pushc( (char *) &doubles[ i ] );
                 }
              else
                 _pushnd( _parnd( i ) );
              break;

         case HB_IT_DATE:
              if ( ISBYREF( i ) )
                 {
                 doubles[ i ] = DateToDbl( _pards( i ) );
                 _pushc( (char *) &doubles[ i ] );
                 }
              else
                 _pushnd( DateToDbl( _pards( i ) ) );
              break;

         default:
              _pushc( NULL );
              break;
         }

   nOleError = pFunc( pUnk );

   for ( i = 3; i <= PCOUNT(); i++ )
      if ( ISBYREF( i ) )
         switch ( (_parinfo( i ) & ~HB_IT_BYREF) )
            {
            case CHARACTER:
            case MEMO:
                 _storc( (char *) _parc( i ), i );
                 break;

            case LOGICAL:
                 _storl( (long) ptros[ i ], i );
                 break;

            case HB_IT_INTEGER:
            case HB_IT_LONG:
                 _stornl( (long) ptros[ i ], i );
                 break;

            case HB_IT_DOUBLE:
                 _stornd( doubles[ i ], i );
                 break;

            case HB_IT_DATE:
                 _stords( DblToDate( doubles[ i ] ), i );
                 break;
            }

   _ret();
}
#endif

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_OLEERROR()
#else
CLIPPER OLEERROR()
#endif
{
   _retnl( (LONG) nOleError );
}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_OLEISOBJECT()
#else
CLIPPER OLEISOBJEC()
#endif
{
   _retl( RetVal.vt == VT_DISPATCH );
}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_OLEUNINITIALIZE()
#else
CLIPPER OLEUNINITI()
#endif
{
   if ( nInitialized > 0 )
      {
      nInitialized--;
      if ( nInitialized == 0 )
         OleUninitialize();
      }
}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_OLE2TXTERROR()
#else
CLIPPER OLE2TXTERR()
#endif
{
   switch ( (LONG) nOleError)
      {
      case S_OK:
         _retc( "S_OK" );
         break;

      case CO_E_CLASSSTRING:
         _retc( "CO_E_CLASSSTRING" );
         break;

      case OLE_E_WRONGCOMPOBJ:
         _retc( "OLE_E_WRONGCOMPOBJ" );
         break;

      case REGDB_E_CLASSNOTREG:
         _retc( "REGDB_E_CLASSNOTREG" );
         break;

      case REGDB_E_WRITEREGDB:
         _retc( "REGDB_E_WRITEREGDB" );
         break;

      case E_OUTOFMEMORY:
         _retc( "E_OUTOFMEMORY" );
         break;

      case E_INVALIDARG:
         _retc( "E_INVALIDARG" );
         break;

      case E_UNEXPECTED:
         _retc( "E_UNEXPECTED" );
         break;

      case DISP_E_UNKNOWNNAME:
         _retc( "DISP_E_UNKNOWNNAME" );
         break;

      case DISP_E_UNKNOWNLCID:
         _retc( "DISP_E_UNKNOWNLCID" );
         break;

      case DISP_E_BADPARAMCOUNT:
         _retc( "DISP_E_BADPARAMCOUNT" );
         break;

      case DISP_E_BADVARTYPE:
         _retc( "DISP_E_BADVARTYPE" );
         break;

      case DISP_E_EXCEPTION:
         _retc( "DISP_E_EXCEPTION" );
         break;

      case DISP_E_MEMBERNOTFOUND:
         _retc( "DISP_E_MEMBERNOTFOUND" );
         break;

      case DISP_E_NONAMEDARGS:
         _retc( "DISP_E_NONAMEDARGS" );
         break;

      case DISP_E_OVERFLOW:
         _retc( "DISP_E_OVERFLOW" );
         break;

      case DISP_E_PARAMNOTFOUND:
         _retc( "DISP_E_PARAMNOTFOUND" );
         break;

      case DISP_E_TYPEMISMATCH:
         _retc( "DISP_E_TYPEMISMATCH" );
         break;

      case DISP_E_UNKNOWNINTERFACE:
         _retc( "DISP_E_UNKNOWNINTERFACE" );
         break;

      case DISP_E_PARAMNOTOPTIONAL:
         _retc( "DISP_E_PARAMNOTOPTIONAL" );
         break;

      default:
         _retc( "Unknown error" );
         break;
      };
}

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_ANSITOWIDE()  // ( cAnsiStr ) -> cWideStr
{
   WORD wLen;
   LPSTR cOut;

   wLen = MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, _parc( 1 ), -1, 0, 0 );
   cOut = ( char * ) _xgrab( wLen * 2 );
   MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, _parc( 1 ), -1,
                        ( LPWSTR ) cOut, wLen );

   _retclen( cOut, wLen * 2 - 1 );
   _xfree( cOut );
}
#endif

//---------------------------------------------------------------------------//

#ifdef __FLAT__
HARBOUR HB_FUN_WIDETOANSI()  // ( cWideStr, nLen ) -> cAnsiStr
{
   WORD wLen;
   LPWSTR cWideStr;
   LPSTR cOut;

   cWideStr = ( LPWSTR ) _parc( 1 );
   wLen = WideCharToMultiByte( CP_ACP, WC_COMPOSITECHECK, cWideStr, -1,
                               cOut, 0, NULL, NULL );
   cOut = ( char * ) _xgrab( wLen );
   WideCharToMultiByte( CP_ACP, WC_COMPOSITECHECK, cWideStr, -1,
                        cOut, wLen, NULL, NULL );

   _retc( cOut );
   _xfree( cOut );
}
#endif

//---------------------------------------------------------------------------//
